from __future__ import annotations

import logging
from http import HTTPStatus
from importlib.metadata import version

from asgiref.typing import HTTPResponseStartEvent, HTTPScope
from httptools import parse_url

from fdk_asgi.exceptions import (
    FnMiddlewareError,
    MethodNotAllowedError,
    MissingMethodError,
    MissingUrlError,
    PathNotFoundError,
)
from fdk_asgi.types import ASGIApp, Message, Receive, Scope, Send
from fdk_asgi.utils import get_client_addr, get_path_with_query_string

FN_FDK_VERSION_HEADER = (
    b"fn-fdk-version",
    f"fdk-asgi/{version(__package__)}".encode(),
)

FN_HTTP_H_ = b"fn-http-h-"
FN_HTTP_REQUEST_URL = b"fn-http-request-url"
FN_HTTP_REQUEST_METHOD = b"fn-http-method"
FN_HTTP_STATUS = b"fn-http-status"
FN_ALLOWED_RESPONSE_CODES = [
    HTTPStatus.OK,
    HTTPStatus.BAD_GATEWAY,
    HTTPStatus.GATEWAY_TIMEOUT,
]

logger = logging.getLogger(__name__)
logger_access = logging.getLogger(f"{__package__}.access")


class FnMiddleware:
    """A pure ASGI middleware, wrapping a regular ASGI application
    and translating Fn <-> REST."""

    def __init__(self, app: ASGIApp, prefix: str = ""):
        self.app = app
        self.prefix = prefix

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # leave all but HTTP connection scopes untouched
        # note that websockets are not supported by fn
        if scope["type"] != "http":
            return await self.app(scope, receive, send)
        scope: HTTPScope

        try:
            mapped_scope = self._map_http_scope(scope)
        except FnMiddlewareError as exception:
            logger.critical(exception)
            await send(
                {
                    "type": "http.response.start",
                    "status": exception.code,
                    "headers": [
                        [b"content-type", b"text/plain"],
                    ],
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": str(exception).encode(),
                }
            )
            return
        await self.app(mapped_scope, receive, self._wrap_send(send, scope))

    def _map_http_scope(self, scope: HTTPScope) -> HTTPScope:
        """Transforms headers etc. sent by Fn/API Gateway
        so that ASGI apps can understand them."""

        # see https://asgi.readthedocs.io/en/latest/specs/www.html#http-connection-scope

        if scope["path"] != "/call":
            raise PathNotFoundError()
        if scope["method"] != "POST":
            raise MethodNotAllowedError()

        http_headers = []
        request_url: bytes | None = None
        request_method: bytes | None = None
        for key, value in scope["headers"]:
            key_lower = key.lower()
            if key_lower.startswith(FN_HTTP_H_):
                http_headers.append((key[len(FN_HTTP_H_) :], value))
            elif key_lower == FN_HTTP_REQUEST_URL:
                request_url = value
            elif key_lower == FN_HTTP_REQUEST_METHOD:
                request_method = value
            else:
                http_headers.append((key, value))

        try:
            parsed_url = parse_url(request_url)
        except TypeError:
            raise MissingUrlError() from None

        if request_method is None:
            raise MissingMethodError()

        scope["method"] = request_method.decode()
        path = parsed_url.path.decode()
        if len(self.prefix) and path.startswith(self.prefix):
            path = path[len(self.prefix) :]
        scope["path"] = path
        scope["raw_path"] = parsed_url.path  # byte-string, excluding any query string
        scope["query_string"] = parsed_url.query or b""  # byte-string
        # scope has precedence over environment variable
        scope["root_path"] = self.prefix
        scope["headers"] = http_headers

        return scope

    @staticmethod
    def _wrap_send(send: Send, scope: HTTPScope) -> Send:
        async def wrapped_send(message: Message):
            # only process messages of type=http.response.start,
            # leave message of other types untouched
            if message["type"] == "http.response.start":
                message: HTTPResponseStartEvent

                new_headers = [
                    (key, value)
                    if key.lower() == b"content-type"
                    else (FN_HTTP_H_ + key, value)
                    for key, value in message["headers"]
                ]
                new_headers.append((FN_HTTP_STATUS, str(message["status"]).encode()))
                new_headers.append(FN_FDK_VERSION_HEADER)

                message["headers"] = new_headers

                original_status_code = message["status"]
                if message["status"] not in FN_ALLOWED_RESPONSE_CODES:
                    message["status"] = HTTPStatus.OK

                logger_access.info(
                    '%s - "%s %s HTTP/%s" %d',
                    get_client_addr(scope),
                    scope["method"],
                    get_path_with_query_string(scope),
                    scope["http_version"],
                    original_status_code,
                )

            await send(message)

        return wrapped_send



import typing
from enum import Enum

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]


class StrEnum(str, Enum):
    def __str__(self):
        return self.value


class HTTPProtocolType(StrEnum):
    auto = "auto"
    h11 = "h11"
    httptools = "httptools"


class LifespanType(StrEnum):
    auto = "auto"
    on = "on"
    off = "off"


class LoopSetupType(StrEnum):
    none = "none"
    auto = "auto"
    asyncio = "asyncio"
    uvloop = "uvloop"



from asgiref.typing import HTTPResponseStartEvent, HTTPScope, Scope
from httptools import parse_url

try:
    from httpx import Headers
except ModuleNotFoundError as exception:  # pragma: no cover
    msg = f"Using {__name__} requires the httpx package to be installed."
    raise RuntimeError(msg) from exception

from fdk_asgi.app import (
    FN_ALLOWED_RESPONSE_CODES,
    FN_HTTP_H_,
    FN_HTTP_REQUEST_METHOD,
    FN_HTTP_REQUEST_URL,
    FN_HTTP_STATUS,
)
from fdk_asgi.types import ASGIApp, Message


class InverseFnMiddleware:
    def __init__(self, app: ASGIApp, *, url_prefix: str = "http://testclient"):
        self.app = app
        self.url_prefix = url_prefix.rstrip("/").encode()

    def _construct_request_url(self, scope: HTTPScope) -> bytes:
        raw_path = scope["raw_path"] or parse_url(scope["path"]).path
        return self.url_prefix + raw_path + scope["query_string"]

    async def __call__(self, scope: Scope, receive, send):
        # leave all but HTTP connection scopes untouched
        # note that websockets are not supported by fn
        if scope["type"] != "http":
            return await self.app(scope, receive, send)
        scope: HTTPScope

        mapped_headers = [
            (FN_HTTP_REQUEST_URL, self._construct_request_url(scope)),
            (FN_HTTP_REQUEST_METHOD, scope["method"].encode()),
        ]
        for raw_key, value in scope["headers"]:
            if raw_key.lower() == b"content-type":
                mapped_headers.append((raw_key, value))
                continue
            mapped_headers.append((FN_HTTP_H_ + raw_key, value))
        scope["headers"] = mapped_headers

        scope["method"] = "POST"
        scope["path"] = "/call"

        return await self.app(scope, receive, self._wrap_send(send))

    @staticmethod
    def _wrap_send(send):
        async def wrapped_send(message: Message):
            # only process messages of type=http.response.start,
            # leave message of other types untouched
            if message["type"] != "http.response.start":
                await send(message)
                return
            message: HTTPResponseStartEvent

            headers = Headers(list(message["headers"]))
            assert "content-type" in headers
            assert "fn-http-h-content-type" not in headers

            assert message["status"] in FN_ALLOWED_RESPONSE_CODES
            assert FN_HTTP_STATUS.decode() in headers
            status_code_header = headers.pop(FN_HTTP_STATUS.decode())
            try:
                message["status"] = int(status_code_header)
            except ValueError as exception:
                msg = f"{FN_HTTP_STATUS.decode()} header cannot be parsed as an integer"
                raise AssertionError(msg) from exception

            for key_lowercase in headers:
                if key_lowercase == "content-type":
                    continue

                assert key_lowercase.startswith("fn-")

            message["headers"] = Headers(
                [
                    (
                        key_raw[len(FN_HTTP_H_) :]
                        if key_raw.lower().startswith(FN_HTTP_H_)
                        else key_raw,
                        value_raw,
                    )
                    for key_raw, value_raw in headers.raw
                ]
            ).raw

            await send(message)

        return wrapped_send



import urllib.parse

from asgiref.typing import WWWScope

# Code taken from uvicorn.protocols.utils

# Copyright © 2017-present, [Encode OSS Ltd](https://www.encode.io/).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def get_client_addr(scope: WWWScope) -> str:
    client = scope.get("client")
    if not client:
        return ""
    # TestClient produces a list instead of a tuple
    # convert to tuple as a workaround
    return "%s:%d" % tuple(client)


def get_path_with_query_string(scope: WWWScope) -> str:
    path_with_query_string = urllib.parse.quote(scope["path"])
    if scope["query_string"]:
        path_with_query_string = "{}?{}".format(
            path_with_query_string, scope["query_string"].decode("ascii")
        )
    return path_with_query_string



from __future__ import annotations

from http import HTTPStatus


class FnMiddlewareError(Exception):
    code: HTTPStatus


class PathNotFoundError(FnMiddlewareError):
    code: HTTPStatus = HTTPStatus.NOT_FOUND

    def __init__(self, msg: str = "Path not found!"):
        super().__init__(msg)


class MethodNotAllowedError(FnMiddlewareError):
    code: HTTPStatus = HTTPStatus.METHOD_NOT_ALLOWED

    def __init__(self, msg: str = "Method not allowed!"):
        super().__init__(msg)


class MissingUrlError(FnMiddlewareError):
    code: HTTPStatus = HTTPStatus.BAD_REQUEST

    def __init__(self, msg: str = "Could not determine request URL!"):
        super().__init__(msg)


class MissingMethodError(FnMiddlewareError):
    code: HTTPStatus = HTTPStatus.BAD_REQUEST

    def __init__(self, msg: str = "Could not determine request method!"):
        super().__init__(msg)



import logging
from pathlib import Path
from typing import Annotated, Optional

import typer
import uvicorn
from uvicorn.importer import import_from_string

from fdk_asgi.app import FnMiddleware
from fdk_asgi.types import HTTPProtocolType, LifespanType, LoopSetupType

UDS_PREFIX = "unix:"
DEFAULT_LOGGING_CONFIG: dict[str, any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "fdk_asgi": {"handlers": ["default"], "level": "INFO"},
        "fdk_asgi.access": {
            "handlers": ["access"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "uvicorn.error": {"level": "INFO"},
    },
}

app = typer.Typer()
logger = logging.getLogger(__name__)


@app.command()
def serve(
    app_uri: Annotated[str, typer.Argument(metavar="APP")],
    uds: Annotated[
        str,
        typer.Option(
            envvar="FN_LISTENER",
            help='Path to the UNIX domain socket, prefixed with "unix:". '
            "This will be managed by the Fn Server.",
        ),
    ] = "unix:./fdk-asgi.socket",
    loop: Annotated[
        LoopSetupType, typer.Option(envvar="FDK_ASGI_LOOP")
    ] = LoopSetupType.none,
    http: Annotated[
        HTTPProtocolType, typer.Option(envvar="FDK_ASGI_HTTP")
    ] = HTTPProtocolType.auto,
    lifespan: Annotated[
        LifespanType, typer.Option(envvar="FDK_ASGI_LIFESPAN")
    ] = LifespanType.auto,
    env_file: Annotated[
        Optional[Path],
        typer.Option(
            envvar="FDK_ASGI_ENV_FILE",
            help="Read configuration from an env file. "
            "Only affects the ASGI app, not the FDK/Server!",
        ),
    ] = None,
    log_config: Annotated[
        Optional[Path], typer.Option(envvar="FDK_ASGI_LOG_CONFIG")
    ] = None,
    log_level: Annotated[
        Optional[str], typer.Option(envvar="FDK_ASGI_LOG_LEVEL")
    ] = None,
    proxy_headers: Annotated[
        bool,
        typer.Option(
            envvar="FDK_ASGI_PROXY_HEADERS",
            help="Enable/Disable X-Forwarded-Proto, X-Forwarded-For, X-Forwarded-Port "
            "to populate remote address info.",
        ),
    ] = True,
    server_header: Annotated[
        bool, typer.Option(envvar="FDK_ASGI_SERVER_HEADER")
    ] = True,
    date_header: Annotated[bool, typer.Option(envvar="FDK_ASGI_DATE_HEADER")] = True,
    prefix: Annotated[
        str,
        typer.Option(
            envvar="FDK_ASGI_PREFIX",
            help="Strips the given prefix from URL paths. "
            "Also sets root_path to this value.",
            show_default=False,
        ),
    ] = "",
    timeout_keep_alive: Annotated[
        int, typer.Option(envvar="FDK_ASGI_TIMEOUT_KEEP_ALIVE")
    ] = 5,
    factory: Annotated[
        bool,
        typer.Option(
            envvar="FDK_ASGI_FACTORY",
            help="Treat APP as an application factory, "
            "i.e. a () -> <ASGI app> callable.",
        ),
    ] = False,
    h11_max_incomplete_event_size: Annotated[
        Optional[int], typer.Option(envvar="FDK_ASGI_H11_MAX_INCOMPLETE_EVENT_SIZE")
    ] = None,
):
    asgi_app = import_from_string(app_uri)
    if factory:
        asgi_app = asgi_app()
    fn_asgi_app = FnMiddleware(asgi_app, prefix)

    socket = Path(uds[len(UDS_PREFIX) :]) if uds.startswith(UDS_PREFIX) else Path(uds)
    # os.umask(0o666)  # todo: check if this is necessary

    config = uvicorn.Config(
        app=fn_asgi_app,
        uds=str(socket),
        loop=loop.value,
        http=http.value,
        ws="none",
        lifespan=lifespan.value,
        env_file=env_file,
        log_config=DEFAULT_LOGGING_CONFIG if log_config is None else log_config,
        log_level=log_level,
        access_log=False,
        # use_colors: Optional[bool] = None,
        interface="asgi3",
        workers=1,
        proxy_headers=proxy_headers,  # todo: check if Functions supports this header
        server_header=server_header,  # todo: check if Functions supports this header
        date_header=date_header,  # todo: check if Functions supports this header
        # forwarded_allow_ips: Optional[Union[List[str], str]] = None,
        # root_path="",
        # limit_concurrency: Optional[int] = None,
        # limit_max_requests: Optional[int] = None,
        # backlog: int = 2048,
        timeout_keep_alive=timeout_keep_alive,
        # timeout_notify: int = 30,
        # timeout_graceful_shutdown: Optional[int] = None,
        # callback_notify: Optional[Callable[..., Awaitable[None]]] = None,
        # headers: Optional[List[Tuple[str, str]]] = None,
        factory=False,
        h11_max_incomplete_event_size=h11_max_incomplete_event_size,
    )

    server = uvicorn.Server(config)
    try:
        server.run()
    finally:
        socket.unlink(missing_ok=True)
